#   author:martinmhan@yahoo.com date:  21/02/2014
#   scadat is a bash command line interface to the tango scada (and in future epics)
#   Copyright (C) <2014>  <Martin Mohan>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software Foundation,
#   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

import PyTango
import nmap
import sys,os
import scadat_base

class scadat_tango(scadat_base.scadat_base):
    """set os.environ['TANGO_HOST'] to point to machine required\n\
1. host=myhost (passed by user)\n\
2. Value returned by environment\n"""
    def __init__(self,verbose=0,host=None):
        super(scadat_tango, self).__init__()
        self.verbose=verbose
#        os.environ['TANGO_HOST']="localhost:10000"
        if host:
            os.environ['TANGO_HOST']=host
        try:
            self.host,self.port=os.environ['TANGO_HOST'].split(":")
        except Exception as e:
            self.handle_exception(e,"set TANGO_HOST e.g. export TANGO_HOST=scadat.org:10000")
#            self.handle_exception(e,"TANGO_HOST="+os.environ['TANGO_HOST']+" Invalid format")
            sys.exit(0)
           
        if(self.IsPortOpen(self.host,self.port)):
            if(self.SetTangoEnviron(self.host,self.port)):
                self.stdout ("open: TANGO_HOST="+os.environ['TANGO_HOST'],2)
        else:
                sys.exit(0)

    def IsPortOpen(self,host,port):
        # Check if host:port is open using nmap
        nm = nmap.PortScanner()
        connect=nm.scan(hosts=host,arguments='-p 10000 -Pn -T5 --unprivileged')
        if not connect['scan']:                
            self.stdout ("Could not connect to "+host+":"+port,1)
            self.stdout ("connect info "+host+":"+port+"->"+str(connect),2)
            return False
        else:
            return True

    def SetTangoEnviron(self,host,port):
        try:
            os.environ['TANGO_HOST']=host+":"+port
            self.db=PyTango.Database()
            return True
        except Exception as e:
            self.handle_exception(e,"FAIL open: TANGO_HOST="+os.environ['TANGO_HOST'])
            return False
#            sys.exit(0)

    def scan(self,network):
        """ scan network using nmap"""
        nm = nmap.PortScanner()
#        nm.scan(hosts=network,arguments='-p 10000')
        nm.scan(hosts=network,arguments='-p 10000 -Pn -T5 --unprivileged')
        hosts_list = [(x,nm[x]['hostname'], nm[x]['tcp'][10000]['state'], nm[x]['tcp'][10000]['name']) for x in nm.all_hosts()]
        for hostip,hostname,state,name in hosts_list:
            self.stdout("hostip "+str(hostip)+" hostname "+hostname+" state "+state+" name "+name,2)
            if(state=="open" and name=="snet-sensor-mgmt"):
                if(self.SetTangoEnviron(hostname,self.port)):
                    serverName=os.environ['TANGO_HOST'].split(":")[0] # simulator/test -> simulator
                    self.stdout (serverName)
                    self.stdout ("opened: TANGO_HOST="+os.environ['TANGO_HOST'],1)
        sys.exit(0)

    def info(self):
        """SCADA specific information on the tango host"""
        self.stdout("scadat_version "+str(self.version))
        self.stdout("TANGO_HOST "+os.environ['TANGO_HOST'])
        self.stdout("TANGO_VERSION - Unknown")
        self.stdout("python.version_info (local) "+str(sys.version))
        self.stdout("PyTango.__version__ "+PyTango.__version__)
        self.stdout(self.db.get_info())

        host_list=self.db.get_host_list()
        MyHosts="\nHosts:"
        for host in host_list:
            MyHosts=MyHosts+("\t%s"%(host))
        self.stdout(MyHosts)

    def ldev(self):
        """List tango devices\n\
-v0 - Starter exported devices\n
-v1 - All exported devices\n
https://www.tango-controls.org/static/PyTango/latest/doc/html/contents.html\n\
        """                
        v=self.verbose
        dev_list=self.db.get_device_exported("*")       # All devices
        for mydev in dev_list:
            self.stdout(mydev,1)
            myclass=self.db.get_class_for_device(mydev)
            if myclass=="Starter" and self.verbose==0:  # Only Starter devices is v==0
                self.get_starter_devices(mydev)
            self.stdout(myclass+" "+mydev,2)

    def open(self,deviceName):
        """Open deviceName and return deviceHandle or error"""
        try:
            DeviceHandle = PyTango.DeviceProxy(deviceName)
            self.stdout("Open device "+deviceName,2)
            return DeviceHandle
        except Exception as e:
            self.handle_exception(e,"Cannot Open device "+deviceName+" on "+self.host+":"+self.port)
#            self.stdout(str(sys.exc_info()[1]))
            sys.exit(0)

#    def read(self,deviceName,attr_name=None):
    def read(self,deviceNames):
        """read device aa/bb/cc e.g. scadat -r aa/bb/cc Amplitude"""
#        v=self.verbose
#        if attr_name:       # read 1 attribute
#            self.attr_read(DeviceHandle,attr_name)

        for deviceName in deviceNames:
#            print ("deviceName=%s"%deviceName)
            DeviceHandle=self.open(deviceName)
            try:
                attr_list=DeviceHandle.get_attribute_list()
                for attr_name in attr_list:
                    self.attr_read(DeviceHandle,attr_name)
            except Exception as e:
                self.handle_exception(e,str(DeviceHandle)+" :Cannot get_attr_list")

    def write(self,deviceName,attr_name=None,value=None):
        """Send command to a tango device in format attr_name(command) arg value"""
        self.stdout("self.write deviceName "+deviceName+" attr_name "+str(attr_name)+" value "+str(value),2)

        DeviceHandle=self.open(deviceName)
        if not attr_name and not value:
            self.stdout("Usage: scadat -w "+deviceName+" <attr_name> <value>")
        elif attr_name and not value:
            attr_value=DeviceHandle.read_attribute(attr_name)
            self.stdout("Usage: scadat -w "+deviceName+"  "+str(attr_name)+" <"+str(attr_value.type)+">")
        else:
            try:         # Try to write attribute:
                self.attr_write(DeviceHandle,attr_name,value)
            except Exception as e:
                self.handle_exception(e,"Error writing cmd "+attr_name+"->"+value)

    def cmd(self,deviceName,cmd_name=None,value=None):
        """Send command to a tango device <command> <value>"""
        self.stdout("self.cmd deviceName "+deviceName+" cmd_name "+str(cmd_name)+" value "+str(value),2)

        DeviceHandle=self.open(deviceName)
        if not cmd_name and not value:
            self.cmd_read_all(deviceName) # also read commands
            self.stdout("Usage: scadat -c "+deviceName+" <cmd_name> <value>")
        elif cmd_name and not value:
            cmd_type=DeviceHandle.command_query(cmd_name)
            if cmd_type.in_type == PyTango.DevVoid:
#                print "Type DevVoid"
                self.cmd_write(DeviceHandle,cmd_name,PyTango.DevVoid)
            else:
                self.stdout("Usage: scadat -c "+deviceName+"  "+str(cmd_name)+" <"+str(cmd_type.in_type)+">")
        else:
            try:         # Try to write command:
                self.cmd_write(DeviceHandle,cmd_name,value)
            except Exception as e:
                self.handle_exception(e,"Error writing cmd "+cmd_name+"->"+value)

    def WriteAttr(self,deviceName,attr_name,value):
        v=self.verbose
        DeviceHandle=self.open(deviceName)
        try:            # try to write attribute            
            self.attr_write(DeviceHandle,attr_name,value)
        except Exception as e:
            self.handle_exception(e,"Error writing attr "+attr_name+"->"+value)

    def attr_write(self,DeviceHandle,attr_name,value):
        """ write 1 attribute """
        v=self.verbose
        attr_value=DeviceHandle.read_attribute(attr_name)
        input_value=self.getValueType(attr_value.type,value)
        try:
            attr_value=DeviceHandle.write_attribute(attr_name,input_value)
        except Exception as e:
            self.handle_exception(e,"error: DeviceHandle.write_attribute(attr_name "+attr_name+",input_value"+str(input_value))
            return

        self.stdout(attr_name+"->"+str(input_value))

    def attr_read(self,DeviceHandle,attr_name):
        """ read 1 attribute """
        v=self.verbose
        try:            # try to read attribute            
            attr_value=DeviceHandle.read_attribute(attr_name)
            self.stdout(str(attr_name))
            self.stdout(str(attr_name)+" "+str(attr_value.value),1)
            self.stdout("attr_name: "+str(attr_name)+" value: "+str(attr_value),2)
        except Exception as e:
            self.handle_exception(e,"Error reading attr "+attr_name)

    def attr_read_all(self,deviceName): # Displays if attribute writeable
        """ read all attributes """
        DeviceHandle=self.open(deviceName)
        v=self.verbose
        try:
            attr_list=DeviceHandle.get_attribute_list()
            for attr_name in attr_list:
                try:
                    attr_value=DeviceHandle.read_attribute(attr_name)
                except Exception as e:
                    self.handle_exception(e,"Cannot read_attribute "+attr_name)

                if attr_value.w_value==None:
                    self.stdout(str(attr_name)+ " <attr> <ro>",1)
                else:
                    self.stdout(str(attr_name)+ " <attr> <w>",1)
        except Exception as e:
            self.handle_exception(e,str(DeviceHandle)+" :Cannot get_attr_list")
#            sys.exit(0)

    def cmd_write(self,DeviceHandle,cmd_name,value):
        """cmd_write and convert value to correct float, int ,string etc.."""
        cmd_type=DeviceHandle.command_query(cmd_name)
        input_value=self.getValueType(cmd_type.in_type,value)
        self.stdout("self.cmd_write cmd_type.in_type "+str(cmd_type.in_type)+" value "+str(value)+" ->input_value "+str(input_value),2)
        cmd_value=DeviceHandle.command_inout(cmd_name,input_value)
        self.stdout(cmd_value)
        self.stdout("DeviceHandle.command_inout: cmd_name "+cmd_name+", input_value "+str(input_value)+" -> cmd_value "+str(cmd_value)+" -> cmd_type.in_type "+str(cmd_type.in_type),2)

    def cmd_read_all(self,deviceName):
        """ read all cmds """
        v=self.verbose
        DeviceHandle=self.open(deviceName)

        try:
            cmd_list=DeviceHandle.command_list_query() # all
            for cmd in cmd_list:
                self.stdout(cmd.cmd_name)
                self.stdout(cmd.cmd_name+" <cmd>"+" in_type: "+str(cmd.in_type),1)
                self.stdout(cmd,2)
        except Exception as e:
            self.handle_exception(e,str(DeviceHandle)+" :Cannot get cmd_list")

    def get_starter_devices(self,device_starter):
        """'Read device Starter e.g. tango/admin/cmpl39)"""
        v=self.verbose

#        self.stdout("get_starter_device: trying to read device_starter "+device_starter+" from "+str(os.environ['TANGO_HOST']),3)
        try:
            deviceHandle = PyTango.DeviceProxy(device_starter)
            RunningServers=deviceHandle.read_attribute("RunningServers").value
            for servers in RunningServers:
                serverName=servers.split("/")[0] # Simulator/test -> Simulator
                self.stdout(serverName,2)
                deviceNames=self.db.get_device_exported_for_class(serverName)
                for deviceName in deviceNames:
                    self.stdout(deviceName)
        except Exception as e:
            self.handle_exception(e,"Error reading "+device_starter+ " from "+self.host+":"+self.port)
            sys.exit(0)

    def read_attribute(self,DeviceHandle,attr): 
        """read device attribute or verbose error"""
        try:
            myvalue=DeviceHandle.read_attribute(attr)
            self.stdout("attr: "+attr)
            self.stdout("attr: "+attr+" value: "+myvalue.value,1)
            self.stdout("attr: "+attr+" value: "+myvalue,2)
        except Exception as e:
            self.handle_exception(e," Unable to read attribute "+attr)


    # need to transform values for writing?
    def getValueType(self,cmd_type,value):
#        self.out("getValueType cmd_type %s value %s"%(cmd_type,value))
        if cmd_type==PyTango.DevFloat:
            return float(value)
        elif cmd_type==PyTango.DevDouble:
            return float(value)
        elif cmd_type==PyTango.DevLong:
            return long(value)
        elif cmd_type==PyTango.DevLong64:
            return long(value)
        elif cmd_type==PyTango.DevULong64:
            return long(value)
        elif cmd_type==PyTango.DevULong:
            return long(value)
        elif cmd_type==PyTango.DevInt:
            return int(value)
        elif cmd_type==PyTango.DevBoolean:
            if value==str(0):
                return False
            else:
                return True
        elif cmd_type==PyTango.DevShort:
            return long(value)
        elif cmd_type==PyTango.DevUShort:
            return long(value)
        elif cmd_type==PyTango.DevString:
            return str(value)
        elif cmd_type==PyTango.DevVarLongArray:
            return map(long, value.split())
        elif cmd_type==PyTango.DevVarLong64Array:
            return map(long, value.split())
        elif cmd_type==PyTango.DevVarULong64Array:
            return map(long, value.split())
        elif cmd_type==PyTango.DevVarFloatArray:
            return map(float, value.split())
        elif cmd_type==PyTango.DevVarDoubleArray:
            return map(float, value.split())
        elif cmd_type==PyTango.DevVarShortArray:
            return map(long, value.split())
        elif cmd_type==PyTango.DevVarUShortArray:
            return map(long, value.split())
        elif cmd_type==PyTango.DevVarStringArray:
            return map(str, value.split())
        elif cmd_type==PyTango.DevVoid:
            return None

# Not sure stuff below
        elif cmd_type==PyTango.DevVarDoubleStringArray:
            return "Not tested yet"
#            return map(str, value.split())

        elif cmd_type==PyTango.DevVarCharArray:
#           "1 2 3 4" -> [1 2 3 4]
            return "Not tested yet"
        elif cmd_type==PyTango.DevVarLongStringArray:
            return "Not tested yet"
        elif cmd_type==PyTango.DevState:
            return "Not tested yet"
        else:
            return "Unknown cmd_type %s"%cmd_type
